// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SourceList.cc	--- implementation of methods in SourceList.h
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-12-02: 1.0.0: - first version
//		UL, 1999-12-17: 1.1.0: - brought in line with selectlib-4.23
//		UL, 2000-11-08: 1.2.0: - delete columnData objects in
//								 SourceList::construct()
//		UL, 2003-02-21: 1.3.0: - adapted to Ds9::toRegion() in imgdisplay-1.37
//								 were GeoRegion object is always composed
//								 of two components (included+excluded)
//		UL, 2003-03-20: 1.4.0: - fixing SPR 2295: SourceList construction
//								 from source list: look for source extent
//								 in columns
//										BOX_SIZE: eboxdetect
//										EXTENT	: ewavelet
//										EXT		: emldetect
//								 and for source counts in
//										SCTS	: eboxdetect/ewavelet
//										SCNTS	: emldetect
//		UL, 2003-05-23: 1.5.0: - added new SourceList constructur to
//								 support new input mode of epicbscalgen
//								 (SCR 63)
//							   - setting of new BS misalignment angles for
//								 the CAL in SourceList constructor works
//								 now via SAS_CCFFILES environment variable
//							   - bug fix in Source::angularSeparation()
//							   - various transparent changes
//
//	AUTHOR:
//		U. Lammers, Thu Dec  2 17:22:32 MET 1999
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/srcdisplay/src/SourceList.cc,v 1.18 2014/01/24 15:03:20 rsaxton Exp $
//
#include <cstdio>
#include <iostream>
#include <unistd.h>
#include <cmath>

#include "common.h"
#include "NAN.h"
#include "errstr.h"
#include "stringutil.h"
#include "DataFilter.h"
#include "SetTableColumn.h"
#include "SelectWCS.h"
#include "SourceList.h"
#include "SasTasks.h"
#include "Ds9.h"
#include "CalOalUtility.h"
#include "Celestial.h"
#include "System.h"
#include "Ccf.h"

#include "CalServer.h"
#include "Boresight.h"

using namespace std;

double Source::angularSeparation(const Source &s) const
{
	return sqrt(CalOalUtility::SQR(cos(_dec)*(_ra - s.ra())) +
				CalOalUtility::SQR(_dec - s.dec())) * RAD_TO_DEG * 3600.;
} // Source::angularSeparation()


ostream &operator<<(ostream &os, const Source &src)
{
	os<<"Source: ra="<<RightAscension(src.ra())
	  <<" dec="<<Declination(src.dec())
	  <<" counts: "<<src.counts();

	return os;
} // operator<<


SourceList::SourceList(Table *sourceTable, double sourceRadius,
					   bool includeSources, ShapeBasics::ShapeColor color)
{
	construct(sourceTable, sourceRadius, includeSources, color);
} // SourceList::SourceList()


SourceList::SourceList(const string &boxListSet, 
					   double sourceRadius,
					   bool includeSources,
					   bool uselabel,
					   bool usePosErr,
					   string posErrCol,
					   double syserr,
					   ShapeBasics::ShapeColor color)
{
	DataSet *set = dataSetServer->open(boxListSet, Dal::Read);
	construct(set->table(0), sourceRadius, includeSources, 
                            uselabel, usePosErr, posErrCol, syserr, color);
	dataSetServer->close(set);
} // SourceList::SourceList()



void SourceList::construct(Table *table, 
						   double sourceRadius,
						   bool includeSources,
						   bool uselabel,
						   bool usePosErr,
						   string posErrCol,
						   double syserr,
						   ShapeBasics::ShapeColor color)
{
	ColumnData *cdra = table->column("RA")->data();
	ColumnData *cddec = table->column("DEC")->data();

	real64	   *ra = cdra->real64Data(),
			   *dec = cddec->real64Data();
	unsigned N = table->rows();

	ColumnData *cdradecerr = 0;
	real32 dfradecerr[N],
		   *radecerr = dfradecerr;
        for (unsigned n=0; n<N; radecerr[n++] = 0); // Initialise to 0

       // Read in position errors if required
        if (usePosErr) {
  	  if (table->hasColumn(posErrCol)) {
		cdradecerr = table->column(posErrCol)->data();
		radecerr = cdradecerr->real32Data();
	  } else {
		errstr<<"There's no "<<posErrCol<<" column present. "
                      <<"Assuming positional errors are zero."
                      << Sas::warning("NoPosnErrorsPresent");
	  }
	}

        cout<<"POSERRCOL: "<<posErrCol<<" "<<radecerr[0]<<endl;

	real32 dfboxsize[N],
		   *boxsize = dfboxsize;
	ColumnData *cdboxsize = 0;
	if (table->hasColumn("BOX_SIZE")) {
		//	eboxdetect src list
		cdboxsize = table->column("BOX_SIZE")->data();
		boxsize = cdboxsize->real32Data();
	} else if (table->hasColumn("EXTENT")) {
		//	ewavelet src list
		cdboxsize = table->column("EXTENT")->data();
		boxsize = cdboxsize->real32Data();
	} else if (table->hasColumn("EXT")) {
		//	emldetect src list
		cdboxsize = table->column("EXT")->data();
		boxsize = cdboxsize->real32Data();
	} else {
		errstr<<"There is no BOX_SIZE/EXTENT/EXT column present. Assuming "
				"all box sizes to be zero."<<Sas::warning("NoBoxSizePresent");
		// Setting all boxsizes to zero
		for (unsigned n = 0; n<N; boxsize[n++] = 0);
	}

	real32 dfcts[N],
		   *cts = dfcts;
	ColumnData *cdcts = 0;
	if (table->hasColumn("SCTS")) {
		//	eboxdetect/emldetect src list
		cdcts = table->column("SCTS")->data();
		cts = cdcts->real32Data();
	} else if (table->hasColumn("SCNTS")) {
		//	ewavelet src list
		cdcts = table->column("SCNTS")->data();
		cts = cdcts->real32Data();
	} else {
		errstr<<"There's no SCTS/SCNTS column present. Assuming all source "
				"counts to be zero."<< Sas::warning("NoSctsPresent");
		// Setting all cts to zero
		for (unsigned n=0; n<N; cts[n++] = 0);
	}

	//	populate src table
	for (unsigned n=0; n<N; ++n) {
		if (cts[n] < 0) {
			errstr<<"Counts for source #"<<n+1<<" is negative: "<<cts[n]
				  <<" - skipping"<<Sas::warning("NegativeSourceCounts");
			continue;
		}

                // Set shape radius based on user choice
                double srtemp = sourceRadius;
                if (usePosErr) 
                   srtemp = sqrt(radecerr[n]*radecerr[n] + syserr * syserr)
                                       /3600.;  // Radius in degrees
                
		FCircle * shape = new FCircle(FPoint(ra[n], dec[n]), srtemp, 
                                              includeSources);
		shape->color(color);
		// Add a label
		if (uselabel)
			shape->label(toString(n+1));

		push_back(Source(ra[n]*DEG_TO_RAD, dec[n]*DEG_TO_RAD,
						 shape,	UNT(cts[n]),
						 // if boxsize[n] is NULL (emldetect) source extent
						 // is insignificant - set to 3 for display purposes
						 !NaN.isSet(boxsize[n]) ? UNT(boxsize[n]) : 3,
						 radecerr[n]));
	}

	delete cdra;
	delete cddec;
	delete cdradecerr;
	delete cdboxsize;
	delete cdcts;

	_region.push_back(Component());
	Component &comp = _region[0];
	N = size();
	for (unsigned n=0; n<N; ++n)
		comp.push_back((*this)[n].shape());
} // SourceList::construct()


SourceList::SourceList(const string &eventListName,
					   const string &xAxis, const string &yAxis,
					   const string &bsCcfName, bool,
					   unsigned binX, unsigned binY,
					   double sourceRadius,
					   unsigned nruns, double likemin,
					   bool runShowDetect)
	: _eventListName(eventListName), _xAxis(xAxis), _yAxis(yAxis),
	  _binX(binX), _binY(binY), _minX(0.), _minY(0.),
	  _sourceRadius(sourceRadius), _nruns(nruns), _likemin(likemin)
{
	errstr<<"Setting new Boresight Ccf: `"<<bsCcfName<<"'"
		  <<Sas::message(Msg::Lib, Msg::Sparse);
	System::setEnv(Ccf::SAS_CCFFILES, bsCcfName);

	//	run attcalc
	SasTasks::attcalc.run(eventListName);

	//	run evselect
	string imageSetName = tmpnam(0);
	SasTasks::evselect.image(eventListName, imageSetName, xAxis, yAxis,
							 binX, binY);

	//	run eboxdetect - and show list in ds9
	string boxSetName = tmpnam(0);
	SasTasks::eboxdetect.detect(imageSetName, boxSetName, nruns, likemin);

	//	show found sources in ds9
	if (runShowDetect)
		SasTasks::srcdisplay.show(imageSetName, boxSetName, sourceRadius);

	SetBlock st(boxSetName);
	construct(st.table(), sourceRadius);
	dataSetServer->close(st.dataSet());

	//	compute number of counts in source circle for later source cross ID
	//	get WCS info from array and event list
	SetBlock evl2(eventListName);
	SelectWCS elXwcs(evl2.table(), xAxis), elYwcs(evl2.table(), yAxis);

	for (unsigned n=0; n<size(); ++n) {
		FPoint p = (*this)[n].shape()->refPoint();
		double r = (*this)[n].shape()->radii()[0];
		//	approximate WCS-TAN projection by linear should be good enough
		//	here
		FCircle circ(FPoint((p.x()-elXwcs.refPixelValue())/elXwcs.delta()+
							elXwcs.refPixel(),
							(p.y()-elYwcs.refPixelValue())/elYwcs.delta()+
							elYwcs.refPixel()), r/fabs(elXwcs.delta()));
		DataFilter filter(circ.toSelectExpression(xAxis, yAxis));
		vector<DataFilter::RowRange> dum;
		(*this)[n].counts(filter.doSelect(*evl2.table(), dum));
	}
	dataSetServer->close(evl2.dataSet());

	errstr<<"Found the following sources for new set of boresight misalignment"
			" angles: \n"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	errstr<<"Src#        Ra/Dec         counts              shape"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	for (unsigned n=0; n<size(); ++n) {
		Source &src = (*this)[n];
		FPoint p = src.shape()->refPoint();
		double r = src.shape()->radii()[0];
		FCircle circ(FPoint((p.x()-elXwcs.refPixelValue())/elXwcs.delta()+
							elXwcs.refPixel(),
							(p.y()-elYwcs.refPixelValue())/elYwcs.delta()+
							elYwcs.refPixel()), r/fabs(elXwcs.delta()));

		RightAscension ra(src.ra());
		Declination dec(src.dec());
		errstr<<n<<' '<<ra<<'/'<<dec<<' '
			  <<src.counts()<<' '
			  <<circ.toSelectExpression(xAxis, yAxis)<<endl
			  <<Sas::message(Msg::Lib, Msg::Silent);
	}

	dataSetServer->remove(imageSetName);
	dataSetServer->remove(boxSetName);
}  // SourceList::SourceList()


SourceList::SourceList(const string &eventListName,
					   const string &xAxis, const string &yAxis,
					   unsigned binX, unsigned binY,
					   double sourceRadius,
					   unsigned nruns, double likemin)
	: _eventListName(eventListName),
	   _xAxis(xAxis), _yAxis(yAxis),
	  _binX(binX), _binY(binY), _minX(0.), _minY(0.),
	  _sourceRadius(sourceRadius), _nruns(nruns),
	  _likemin(likemin)
{
	//	run evselect
	string imageSetName = tmpnam(0);
	SasTasks::evselect.image(eventListName, imageSetName, xAxis, yAxis,
							 binX, binY);

	//	run eboxdetect
	string boxSetName = tmpnam(0);
	SasTasks::eboxdetect.detect(imageSetName, boxSetName, nruns, likemin);

	//	show found sources in ds9
	SasTasks::srcdisplay.show(imageSetName, boxSetName, sourceRadius, false);

    cout<<"\nPlease select the celestial sources that you want to use in the search for the boresight misalignment angles. For each source please:\n\n";
	cout<<"  1. Double click on the circular marker around the source.\n";
	cout<<"  2. In the dialog box which comes up, change the marker status from \"Excluded\" to \"Included\"\n";
	cout<<"  3. In the \"Text\" field specify the true celestial coordinates of the source in the form:\n";
	cout<<"        23h59m48.12s / 45o38'14.55\"\n";
	cout<<"When you are done, hit any key and the marked sources and their coordinates will be read.\n";

    char ch;
    cin.get(ch);
	cout<<"Ok ...\n";

	//	region in image coordinates
	GeoRegion r = Ds9::toRegion(SasTasks::imgdisplay.getRegions());
	Component &c = r[0];	//	only the included components
	for (unsigned n=0; n<c.size(); ++n)
		if (c[n]->isLabeled()) {
			vector<string> v(breakup(c[n]->label(), "/#;:"));
			if (v.size() != 2)
				errstr<<"Coordinates for source #"<<n<<" are not "
						"in the right format"
					  <<Sas::error("WrongSourceCoordinates");
			RightAscension ra(v[0]);
			Declination dec(v[1]);
			push_back(Source(ra.angle(), dec.angle(), c[n], 0, 1, 0.));
		} else
			errstr<<"No coordinates specified for source #"<<n
				  <<Sas::error("NoSourceCoordinates");

	completeSourceList(eventListName, imageSetName);
	diagnostic();

	dataSetServer->remove(imageSetName);
	dataSetServer->remove(boxSetName);
} // SourceList::SourceList()


void SourceList::completeSourceList(const string &eventListName,
									const string &imageSetName)
{
    if (size() < 2)
        errstr<<"At least 2 sources are needed in the boresight misalignment "
                "angle computation - only "<<size()<<" have been "
                "specified."<<Sas::error("TooFewSources");
    //  compute number of counts in source circle for later source cross ID
    //  get WCS info from array and event list
    SetBlock eventList(eventListName);
    SelectWCS elXwcs(eventList.table(), _xAxis),
			  elYwcs(eventList.table(), _yAxis);
    SetBlock image(imageSetName);
    SelectWCS imgXwcs(image.array(), 1), imgYwcs(image.array(), 2);
    dataSetServer->close(image.dataSet());
    _minX = elXwcs.refPixel() +
                        (imgXwcs(1)-elXwcs.refPixelValue())/elXwcs.delta(),
    _minY = elYwcs.refPixel() +
                        (imgYwcs(1)-elYwcs.refPixelValue())/elYwcs.delta();
    for (unsigned n=0; n<size(); ++n) {
        DataFilter filter((*this)[n].shape()->
                        toSelectExpression(_xAxis, _yAxis,
									   DBL(_binX), DBL(_binY), _minX, _minY));
        vector<DataFilter::RowRange> dum;
        (*this)[n].counts(filter.doSelect(*eventList.table(), dum));
    }
    dataSetServer->close(eventList.dataSet());
} // SourceList::completeSourceList()


void SourceList::diagnostic() const
{
	errstr<<"The following sources will be used in the computation of the\n"
			"boresight misalignment angles:\n"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	errstr<<"Src#        Ra/Dec         counts              shape"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	for (unsigned n=0; n<size(); ++n) {
		const SourceList &sl = *this;
		RightAscension ra(sl[n].ra());
		Declination dec(sl[n].dec());
		errstr<<n<<' '
			  <<ra<<'/'<<dec<<' '
			  <<sl[n].counts()<<' '
			  <<sl[n].shape()->toSelectExpression(_xAxis, _yAxis,
									DBL(_binX), DBL(_binY), _minX, _minY)<<endl
			  <<Sas::message(Msg::Lib, Msg::Silent);
	}
} // SourceList::diagnostic()


SourceList::SourceList(const std::string &eventListName,
					   const std::string &refSrcListName,
					   const std::string &xAxis, const std::string &yAxis,
					   unsigned binX, unsigned binY,
					   double sourceRadius,
					   unsigned nruns, double likemin)
	: _eventListName(eventListName),
	   _xAxis(xAxis), _yAxis(yAxis),
	  _binX(binX), _binY(binY), _minX(0.), _minY(0.),
	  _sourceRadius(sourceRadius), _nruns(nruns),
	  _likemin(likemin)
{
	//	run evselect
	string imageSetName = tmpnam(0);
	SasTasks::evselect.image(eventListName, imageSetName, xAxis, yAxis,
							 binX, binY);

	//	run eboxdetect
	string boxSetName = tmpnam(0);
	SasTasks::eboxdetect.detect(imageSetName, boxSetName, nruns, likemin);

	//	show found sources in ds9
	SasTasks::srcdisplay.show(imageSetName, boxSetName, sourceRadius, false);

	//	overlay sources from reference source list on displayed image
	SasTasks::srcdisplay.show("", refSrcListName, sourceRadius, false, true,
							  true, ShapeBasics::YELLOW);

	cout<<"The displayed image shows the result of an eboxdetect run on the event list\n"<<"`"<<eventListName<<"' - each found source is marked with a white circular marker with status \"Excluded\". Also shown are the positions of the reference source from the given source list in `"<<refSrcListName<<"' as circular excluded markers in the color yellow. Those markers are also labeled with the index of the respective source starting from 1. Please select now the sources that shall be used in the search for the optimum boresight misalignment angles. For each of the yellow-marked sources please:\n\n"
		"  1. Identify the corresponding source in the set of white-marked sources - please note: It is possible that a yellow source marker overlaps a white marker or vice versa. In this case the yellow marker may be arbitrarily shifted such that the white marker can be selected. Changing the stacking order of overlapping region markers in ds9 can also be done with \"Region\"->\"Move to Front/Back\"\n"
		"  2. Double-click on the corresponding white circle marker around the source\n"
		"  3. In the dialog box which comes up, change the marker status from \"Excluded\" to \"Included\"\n"
		"  4. In the \"Text\" field specify the source index as a single-digit integer number starting from 1.\n\n"
		"When done, hit any key and the marked \"Included\" source and their respective celestial coordinates from the reference source list shall be read.\n";

    char ch;
    cin.get(ch);
	cout<<"Ok ...\n";

	//	get reference source list
	SourceList refSrcList(refSrcListName, sourceRadius, true, false);
	unsigned N = refSrcList.size();

	//	get select sources from image
	GeoRegion r = Ds9::toRegion(SasTasks::imgdisplay.getRegions());
	Component &c = r[0];	//	only the included components
	for (unsigned n=0; n<c.size(); ++n)
		if (c[n]->isLabeled()) {
			unsigned indx = 0;
			try {
				indx = toNumber<unsigned>(c[n]->label());
			} catch(...) {
				continue;
			}
			if (indx > 0 && indx <= N)
				push_back(Source(refSrcList[indx-1].ra(),
								 refSrcList[indx-1].dec(),
								 c[n], 0, 1, 0.));
			else
				errstr<<"Source index "<<indx<<" is out of valid bounds: [1, "
					  <<refSrcList.size()<<']'
					  <<Sas::error("InvalidSourceIndex");
		} else
			errstr<<"No source index specified for source #"<<n
				  <<Sas::error("NoSourceIndex");

	completeSourceList(eventListName, imageSetName);
	diagnostic();

	dataSetServer->remove(imageSetName);
	dataSetServer->remove(boxSetName);
} // SourceList::SourceList()


SourceList::~SourceList()
{
	_region.free();
} // SourceList::~SourceList()


ostream &operator<<(ostream &os, const SourceList &sl)
{
	os<<"Source list:\n";
	SourceList::const_iterator i = sl.begin();
	for (; i != sl.end(); ++i)
		os<<*i<<endl;

	return os;
} // operator<<
